#include "h5augjpss.h"
#include "h5aj_private.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

/* Global variables definitions */
const char    *xmlfilename_g = NULL;
const char    *datafilename_g = NULL;
const char    *hide_unknown_g = NULL;
const char    *update_from_xml_g = NULL;
const char    *update_geoloc_g = NULL;
const char    *flatten_g = NULL;
const char    *restore_links1_g = NULL;
const char    *restore_links4_g = NULL;
const char    *add_links_idv_g = NULL;
char          xmlfile_path_g[1024];
char          datafile_path_g[1024];

void
usage(void)
{
    fprintf(stderr, "\n");
    fprintf(stderr, "usage:\n");
    fprintf(stderr, "%s [-h | --help]\n", progname_g);
    fprintf(stderr, "%s [-on | -ln] [<xmlfile>] <h5file>\n", progname_g);
    fprintf(stderr, "       (-o2 requires xmlfile)\n"); 
    fprintf(stderr, "%s [-rn ] <h5file>\n", progname_g);
    fprintf(stderr, "%s <xmlfile> <h5file>", progname_g);
    fprintf(stderr, " (default - invokes options -o1 -o2 -o3 -o4)\n");
    fprintf(stderr,
        "\n" \
        "REQUIREMENTS\n" \
        "   All working options require the name of a JPSS Product file as the final\n" \
        "   command line argument to the h5augjpss command.  If the argument includes a\n" \
        "   full or relative path to the file, the file in that location will be used.\n" \
        "   If the argument is the filename only and H5AUGJPSS_DATADIR is defined in the\n" \
        "   environment, $H5AUGJPSS_DATADIR will be searched for the file followed by the\n" \
        "   current working directory.\n"
        "   Option -o2 requires the name or path of an JPSS XML product file as the\n" \
        "   second to last argument, with the locations searched in the same order.\n" \
        "   Option -o3 requires an JPSS geolocation product file in the same location\n" \
        "   as the JPSS product file.  The name of this file is specified in an N_GEO_Ref\n" \
        "   attribute in the JPSS product file.\n" \
        "   The default command with no options requires all of these files.\n\n" \
        "FURTHER INFORMATION\n" \
        "   For more detailed information regarding this tool, please see documentation in\n" \
        "   http://www.hdfgroup.org/projects/jpss.\n" \
        "   Please contact help@hdfgroup.org with any questions regarding this tool.\n\n" \
        "EXAMPLE COMMANDS\n" \
        "   %s -o1 <JPSS product file name>\n" \
        "      removes the link to the group with objects unknown to netCDF-4, making the\n" \
        "      group unavailable to the HDF5 library and allowing netCDF-4 based tools\n" \
        "      and applications to open <JPSS product file name>\n" \
        "   %s -o2 <JPSS XML product file name> <JPSS product file name>\n" \
        "      extracts dimensions and other information from the JPSS XML product file\n" \
        "      according to the specifications found in h5augjpss-design.pdf.\n" \
        "   %s -o3 <JPSS product file name>\n" \
        "      checks JPSS product file for Height, Latitude and Longitude datasets\n" \
        "      and copies them from the file specified in the product file's attribute\n" \
        "      N_GEO_Ref if they are not present in the product file.\n" \
        "   %s -o4 <JPSS product file name>\n" \
        "       creates links to the JPSS product file's datasets in /All_Data/...,\n" \
        "       then removes the link to the group /All_Data/..., making it unavailable\n" \
        "       to the HDF5 library and allowing tools such as IDV that don't support\n" \
        "       HDF5 groups to open the modified file.\n" \
        "   %s -o1 -o2 -o3 -o4 <JPSS XML product file name> <JPSS product file name>.\n" \
        "       and\n" \
        "   %s <JPSS XML product file name> <JPSS product file name>\n" \
        "       are equivalent, executing options -o1 -o2 -o3 -o4 as described above.\n" \
        "   %s -r1 <JPSS product file name>\n" \
        "       reverses the effect of -o1, replacing the original link to restore access\n" \
        "       to the /Data_Products group.\n" \
        "   %s -r4 <JPSS product file name>\n" \
        "       reverses the effect of -o4, removing the added links and replacing the\n" \
        "       original link to restore access to the /All_Data group.\n\n" \
        "OPTIONS\n" \
        "   -h, --help\n" \
        "      Print this usage message and exit\n" \
        "   -V, --version\n" \
        "      Print version number and exit\n" \
        "   -o, -l<1 - 4>\n" \
        "      1   remove the link to the group with objects unknown to netCDF-4\n" \
        "      2   update JPSS product file with dimensions and other attribute information\n" \
        "          from the JPSS XML product file according to the specifications in\n" \
        "          h5augjpss-design.pdf document.\n" \
        "      3   update JPSS product file  with geolocation information from file specified\n" \
        "          in its N_GEO_Ref attribute, unless the datasets are already in the file.\n" \
        "      4   create links in \"/\" to datasets in /All_Data/<subgroup>\n" \
        "          and remove the link to the /All_Data group.  This option combined \n" \
        "          with -l 1 removes access to all groups in the file. \n" \
        "   -r<1 or 4>\n" \
        "      1   restore link to the /Data_Products group in the JPSS product file.\n" \
        "      4   remove links in \"/\" to datasets originally in /All_Data/<subgroup>\n" \
        "          and restore the link to the /All_Data group in the JPSS product file. \n\n" \
        "          Options 1 through 4 combined  hide all group structures in the file and\n" \
        "          add dimension scales and geolocation datasets that allow IDV and other\n" \
        "          tools to open the file when certain attributes are added (with h5edit\n" \
        "          or hdfview) to make the file CF compliant.\n\n" \
        "NOTE that options are provided to reverse the modifications in levels 1 and 4\n" \
        "while no option to reverse the modifications in levels 2 or 3 is provided.\n\n" \
        "       \n", progname_g , progname_g, progname_g, progname_g, progname_g, progname_g, progname_g, progname_g);
    fprintf(stderr, "\n");
} /* usage() */

/* Print h5augjpss version */
void
print_version(void)
{
    fprintf(stdout, "%s\n", H5AUGJPSS_VERSION);
}
 
void
leave(int ret){
    /* about to leave the tool. Do some necessary cleanup and then quit. */

    /* quit */
    exit(ret);
}

/* Parse command line options.
 */
void
parse_options(int argc, const char **argv)
{
    int           argno;
    const char    *s = NULL;
    const char    *default_action = "enabled";
    char          *h5augjpss_datadir = getenv("H5AUGJPSS_DATADIR");
    unsigned int  xml_set = 0;
        
 
    if (!h5augjpss_datadir)
        h5augjpss_datadir = "";
    
    /* parse in options */
    for (argno = 1; argno < argc && '-' == argv[argno][0]; argno++){
        if (streq(argv[argno], "--help") || streq(argv[argno], "-h")){
            default_action = NULL;
            usage();
            leave(EXIT_SUCCESS);
        } else if (streq(argv[argno], "--version") || streq(argv[argno], "-V")){
            print_version();
            leave(EXIT_SUCCESS);
        } else if ('o' == argv[argno][1] || 'l' == argv[argno][1]){
            if('1' == argv[argno][2]) {
                default_action = NULL;
                hide_unknown_g = "yes";
            }else if('2' == argv[argno][2]) {
                default_action = NULL;
                update_from_xml_g = "yes";
            }else if('3' == argv[argno][2]) {
                default_action = NULL;
                update_geoloc_g = "yes";
            }else if('4' == argv[argno][2]) {
                default_action = NULL;
                flatten_g = "yes";
            }
        } else if ('r' == argv[argno][1]) {
            if('1' == argv[argno][2]) {
                default_action = NULL;
                restore_links1_g = "yes";
            }else if('4' == argv[argno][2]) {
                default_action = NULL;
                restore_links4_g = "yes";
            }
        } else {  /* unknown command line option */
            fprintf(stderr, "Unknown command line option %s\n", argv[argno]);
            usage();
            leave(EXIT_FAILURE);
        }
    }

    if(default_action) {
        hide_unknown_g = "yes";
        update_from_xml_g = "yes";
        update_geoloc_g = "yes";
        flatten_g = "yes";
    }
        
    if (argno >= argc) {
        fprintf(stderr, "%s: missing datafile argument(s)\n", progname_g);
        usage();
        leave(EXIT_FAILURE);
    }
    
    /* check for xml file - last file must be hdf5 datafile.  If 2 args remaining,
     * assume next is xml file. */  
    if (argno < (argc - 1)) 
        xmlfilename_g = argv[argno++];
    datafilename_g = argv[argno++];
    
    /* if datafilename_g has '/'s, assume it's an absolute path.  If it can't be
     * accessed, report error and exit.  If it's just a filename, and 
     * h5augjpss_datadir isn't empty, check to see if h5augjpss_datadir/datafilename_g 
     * can be accessed.  If it can, assign it to datafilename_g.  If it can't,
     * check datafilename_g and report error if it fails, use it if it succeeds. 
     */
   
    if (!strchr(datafilename_g, '/')) {
        if (h5augjpss_datadir != "") {
            strncpy(datafile_path_g, h5augjpss_datadir, strlen(h5augjpss_datadir));
            if (datafile_path_g[strlen(h5augjpss_datadir)-1] != '/') 
                strncat(datafile_path_g, "/", 1);  
            strncat(datafile_path_g, datafilename_g, strlen(datafilename_g));
        }
        if (access(datafile_path_g, W_OK) == 0) {
            datafilename_g = datafile_path_g;
        }  else if (access(datafilename_g, W_OK) == -1) {
            fprintf(stderr, "Can't access data file %s.\n", datafilename_g);
            leave(EXIT_FAILURE);
        }
    } else if (access(datafilename_g, W_OK) == -1) {
        fprintf(stderr, "Can't access data file %s.\n", datafilename_g);
        leave(EXIT_FAILURE);
    }
    /* At this point, either datafilename_g was an absolute path that can be 
     * accessed (for writing), or it was just a filename which is in the h5augjpss_datadir
     * directory and can be accessed, it was just a filename that is in the current
     * working directory and can be accessed, or we've printed the error message and 
     * exited.
     */     

    if (update_from_xml_g) {
        if (!xmlfilename_g) {
        fprintf(stderr, "\nError:  level 2 requires xmlfile.\n");
            usage();
            leave(EXIT_FAILURE);
        }
        /* if xmlfilename_g is a not a path, ie doesn't have '/'s, check in directory specified in environment variable H5AUGJPSS_DATADIR. */
        if (!strchr(xmlfilename_g, '/')) {
            if (strcmp(h5augjpss_datadir, "") != 0) {
                strncpy(xmlfile_path_g, h5augjpss_datadir, strlen(h5augjpss_datadir));
                strncat(xmlfile_path_g, xmlfilename_g, strlen(xmlfilename_g));
                if (access(xmlfile_path_g, R_OK) == 0) {
                    xmlfilename_g = xmlfile_path_g;
                    xml_set = 1;
                }
            }
            if (!xml_set && access(xmlfilename_g, R_OK) == 0) {
                xml_set = 1;
            }
            if (!xml_set) {              
                /*if datafilename_g is a path, check for xmlfilename_g in that directory */
                char *last_slash = strrchr(datafilename_g, '/'); 
                if (last_slash) {
                    strncpy(xmlfile_path_g, datafilename_g, last_slash - datafilename_g +1);
                    xmlfile_path_g[last_slash - datafilename_g +1] = '\0';
                    strncat(xmlfile_path_g, xmlfilename_g, strlen(xmlfilename_g));  
                    if (access(xmlfile_path_g, R_OK) == 0) { 
                        xmlfilename_g = xmlfile_path_g; 
                        xml_set = 1;                        
                    } else {
                        fprintf(stderr, "Can't access(2) %s.\n", xmlfile_path_g);
                        leave(EXIT_FAILURE);
                    }                    
                }            
            }
        } else if (access(xmlfilename_g, R_OK) == -1) {
            fprintf(stderr, "Can't access(1) %s.\n", xmlfilename_g);
            leave(EXIT_FAILURE);
        }
    }
    if (NULL != xmlfilename_g) 
        fprintf(stderr, "XML file is %s.\n", xmlfilename_g);
    fprintf(stderr, "Data file is %s.\n", datafilename_g); 
}


#define RANK   1

int
h5aj_augment_file() {

    hid_t       file = (-1), group = (-1), rt_group = (-1);         /* Handles */
    herr_t      status;
    htri_t      file_status;
    int         i, j;
    H5ajpdPtr   h5ajpdptr;
    char        *need_to_redo_flatten = NULL;	
/*    char        *level_2_executed = NULL; */
    char        *group_with_reference_types = "/Data_Products";
    char        subgroup_name[1024];

    if(hide_unknown_g || update_from_xml_g || update_geoloc_g || flatten_g) {
        file_status = H5Fis_hdf5(datafilename_g);
        if (file_status == 0) {
            fprintf(stderr, "%s is not an HDF5 file.\n", datafilename_g);
            return -1;
        } else if (file_status < 0) {
            fprintf(stderr, "%s doesn't exist or cannot be accessed.\n", datafilename_g);
            return -1;
        }
        file = H5Fopen(datafilename_g, H5F_ACC_RDWR, H5P_DEFAULT);
        if (file < 0) {
            fprintf(stderr, "Error opening HDF5 data file %s!\n", datafilename_g);
            return -1;
        }
        rt_group = H5Gopen(file, "/", H5P_DEFAULT);


        if(hide_unknown_g) {
            if (0 < H5Lexists(file, group_with_reference_types, H5P_DEFAULT)) {
                fprintf(stderr, "Disconnect Data_products group.\n");
                h5aj_disconnect_group(file, group_with_reference_types);
            } else {
                fprintf(stderr, "%s not found (already disconnected?)\n", group_with_reference_types);
                /* could check for name and address attributes */
            }
        }

	if (update_from_xml_g || update_geoloc_g) {
            if ( 0 >= ( status = H5Lexists(rt_group, "All_Data",  H5P_DEFAULT))) { 
                status = h5aj_restore_group_link(file, "All_Data");
                if (status < 0) {
                    fprintf(stderr, "Error:  link to /All_Data group could not be restored.\n");
                    goto done;
                } else {
                    status = h5aj_get_All_Data_sole_subgroup(file, &group, subgroup_name);
                    if (status < 0) {
                        fprintf(stderr, "Error:  No child group of /All_Data found.\n");
                        goto done;
                    } else {
                        h5aj_remove_links_to_group_objects(file, group);
                        need_to_redo_flatten = "yes";
                    }
                }
            } else {
	        status = h5aj_get_All_Data_sole_subgroup(file, &group, subgroup_name);
	        if (status < 0) {
                   fprintf(stderr, "No child group of All_Data found.\n");
	           goto done;
                }
            }
        }
        if(update_from_xml_g) {
            fprintf(stderr, "Parse xml file %s.\n", xmlfilename_g);
            h5ajpdptr = parseJPSSDataProductFile(xmlfilename_g);

            if(h5ajpdptr) {
                if(strncmp(h5ajpdptr->collectionShortName, subgroup_name , strlen(h5ajpdptr->collectionShortName)) != 0) {
                    fprintf(stderr, "Product profile's collection short name %s doesn't\ncorrespond to All_Data subgroup name %s.", h5ajpdptr->collectionShortName, subgroup_name);       
                    goto done;
                }   
                for (i = 0; i < h5ajpdptr->nproductDatas; i++) {
                    if (h5ajpdptr->productDatas[i] == NULL)
                        return;
    
                    
                    fprintf(stderr, "Process fields.\n");
                    for (j = 0; j < h5ajpdptr->productDatas[i]->nFields; ++j) {
                        h5aj_add_dimension_scales(group, h5ajpdptr->productDatas[i]->fields[j]);
                        h5aj_add_attributes_for_datums(group, h5ajpdptr->productDatas[i]->fields[j]); 
                    }
                }
            }             
        }

        if(update_geoloc_g) {
                fprintf(stderr, "Add geolocation datasets.\n");
                h5aj_add_geolocation_datasets(file, group, datafilename_g);
        }

        if(flatten_g || need_to_redo_flatten) {
            if ( 0 >= ( status = H5Lexists(rt_group, "All_Data",  H5P_DEFAULT))) { 
                fprintf(stderr, "All_Data group can't be found (previously disconnected by executing option\n");
                fprintf(stderr, "-o4?)\n");
                goto done;
            } else {
                status = h5aj_get_All_Data_sole_subgroup(file, &group, subgroup_name);
                if (status < 0) {
                    fprintf(stderr, "Error:  No child group of /All_Data found.\n");
                    goto done;
                }
                fprintf(stderr, "Add links in '/' to datasets in /All_Data/%s, then remove\naccess to /All_Data group.\n", subgroup_name);
                h5aj_link_group_objects(file, group);
            }
        }

 done:
        if(group > 0) 
            status = H5Gclose(group);
        if(rt_group > 0) 
            status = H5Gclose(rt_group);
        if(file> 0)
            status = H5Fclose (file);
   
    } else if (restore_links1_g || restore_links4_g) {
        if (restore_links1_g) {
            file = H5Fopen(datafilename_g, H5F_ACC_RDWR, H5P_DEFAULT);

            h5aj_restore_group_link(file, "Data_Products");

        }   
        if (restore_links4_g) {
            file = H5Fopen(datafilename_g, H5F_ACC_RDWR, H5P_DEFAULT);

            status = h5aj_restore_group_link(file, "All_Data");
            if (status < 0)
                fprintf(stderr, "Error:  link to /All_Data group could not be restored.\n");
            else { 
                status = h5aj_get_All_Data_sole_subgroup(file, &group, subgroup_name);
                if (status < 0) {
                    fprintf(stderr, "Error:  No child group of /All_Data found.\n");
                /*} else if (status <= 800) {*/
                } else {
                    h5aj_remove_links_to_group_objects(file, group);
                }
            }
        }
         
        if(group > 0) 
            status = H5Gclose(group);
        if(file > 0)
            status = H5Fclose(file);
    }
    

    return 0;
}

